# -*- coding: utf-8 -*-

# Le but de ce programme est de présenter des tests non constructifs de primalité. On veut tester si un nombre est 
# composé sans avoir à le factoriser.

import numpy as np

# Dans un premier temps, construisons un crible d'Eratosthène

def crible_Eratosthene (N) :
    is_prime = [True for n in range(N+1)]
    is_prime[0] = False
    is_prime[1] = False
    for n in range(2, N+1) :
        if is_prime[n] :
            for k in range (2*n, N+1, n) :
                is_prime[k] = False
    return [p for p in range(N+1) if is_prime[p]]

print(crible_Eratosthene(100))


# Ensuite, voici un test de primalité paresseux.
# Rappelons qu'un entier n est premier si et seulement si k**(n-1) = 1 mod n pour tout entier 1 <= k < n.
# En particulier, si n >= 2 et 2**(n-1) != 1 mod n, alors n n'est pas premier. On peut ainsi éliminer beauocup de nombres premiers.

# Au passage, remarquons que l'expression 'pow(2, n-1, n) == 1' ci-dessous est un booléen : la fonction 'test_primalite_paresseux' 
# retourne True ou False. Cette syntaxe est plus souvent utilisée dans les boucles conditionnelles (if ...).

def test_primalite_paresseux (n) :
    return pow(2, n-1, n) == 1

def crible_paresseux (N) :
    return [2] + [p for p in range(3, N+1) if test_primalite_paresseux(p)]

print(crible_paresseux(100))


# Ce test a l'avantage de pourvoir s'appliquer à de très grands nombres, sans avoir à calculer leur factorisation.

p_1 = 422341310396484421031508597853  # Un grand nombre premier (30 chiffres)
p_2 = 977185300400297951195163250231  # Un autre grand nombre premier (30 chiffres)

print(test_primalite_paresseux(p_1), test_primalite_paresseux(p_2), test_primalite_paresseux(p_1*p_2))


# Si ce test "paresseux" affirme qu'un nombre n'est pas premier, c'est vrai. Cependant, il peut faire l'erreur opposée, 
# qui est de ne pas remarquer qu'un nombre est composé. Pour ne pas faire cette erreur, il faudrait tester tous les entiers 1 <= k < n, 
# et non seulement k = 2.

# Il n'y a que 3 entiers <= 1000 pour lesquels cela arrive : 341 = 11*31, 561 = 3*11*17, 645 = 3*5*43
# Remarquons que :
# * phi(341) = 10*30 = 300. L'ordre de 2 dans (Z/341Z)* est 10, qui divise bien 300, mais qui divise aussi 340 (pas de chance !).
# * phi(561) = 2*10*16 = 320. L'ordre de 2 dans (Z/561Z)* est 40, qui divise bien 320, mais qui divise aussi 560 (pas de chance !).
# * phi(645) = 2*4*42 = 336. L'ordre de 2 dans (Z/645Z)* est 28, qui divise bien 336, mais qui divise aussi 644 (pas de chance !).

# Ce test n'est pas si mauvais, mais n'est pas parfait. Pour le reforcer, on peut tester d'autres entiers que 2. Une façon de faire 
# consiste à tester des entiers au hasard. Par exemple, en testant 10 entiers au hasard :

def test_primalite_Fermat (n) :
    for t in range(10) :
        k = np.random.randint(1, n)
        if pow(k, n-1, n) != 1 :
            return False
    return True

def crible_Fermat (N) :
    return [p for p in range(2, N+1) if test_primalite_Fermat(p)]

# Ce test est probabiliste, mais a de bonnes chances de réussir.
# Il y a certains nombres (nombres de Carmichael), plutôt rares, qui se "comportent mal" vis-à-vis de ce test. C'est le cas de 561.
# En-dehors de ces nombres, les chances que le test se trompent sont d'au plus 2**(-10) si l'on teste 10 entiers au hasard.
# En testant plus d'entiers, on peut facilement faire descendre cette probabilité afin d'augmenter la fiabilité du test.

# Les nombres de Charmichael restent un obstacle à cet algorithme naïf.
# D'autres tests existent, plus sophistiqués, qui évitent cet écueil : Miller-Rabin, Lucas, Baillie-PSW...